/******************************************************************************
* 	ref18_01/common/iscsi_common.h
*
* 	defines stuff common for both iscsi initiator and iscsi target
*
*   IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
*   By downloading, copying, installing or using the software you agree to this
*   license. If you do not agree to this license, do not download, install,
*   copy or use the software.
*
*   License Agreement
*
*   Copyright (C) 2001  Inter-Operability Lab.(IOL),Univ. of New Hampshire(UNH),
*                       Durham, NH 03824
*
*   Redistribution and use in source and binary forms, with or without
*   modification, are permitted provided that the following conditions are met:
*
*   -Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
*
*   -Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
*
*   The name of IOL/UNH may not be used to endorse or promote products derived
*   from this software without specific prior written permission.
*	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
*   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
*   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL IOL,UNH OR CONTRIBUTORS
*   BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF
*   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
*   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*   POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************/

#ifndef _ISCSI_COMMON_H
#define _ISCSI_COMMON_H

#include <asm/types.h>
#include <linux/inet.h>

#include "debug.h"

#define ISCSI_PORT 			3260		/* IANA's WKP */
#define ISCSI_HDR_LEN			48		/* Basic Header Len */
#define ISCSI_CDB_LEN			16		/* CDB len within BHS */
#define CRC_LEN				4		/* Digest length */
#define MAX_KEY_NAME_LENGTH		63		/* max len of key name*/
#define MAX_KEY_VALUE_LENGTH		255		/* max len of key val */
#define MAX_ISCSI_NAME_LENGTH		223		/* max len of iSCSI name */
#define ALL_ONES			0xFFFFFFFF	/* Reserved Value */
#define ISCSI_DATA_MAX			0xFFFFFFFF
#define QUEUE_DEPTH_ALLOWED		100
#define MAX_SOCK_ADDR      		128
#define SERIAL_BITS             	31

#define WHITE_SPACE			" \t\v\f\n\r"
#define TARGETNAME_HEADER		"iqn.2002-10.edu.unh.iol.iscsi.draft18-target:"

/* max number of loops allowed during login negotiation */
#define LOOP_TIMES			8

/* values for "role" and "who_called" parameters */
#define INITIATOR			1
#define TARGET				2
#define MANAGEMENT			4

#define TRUE				1
#define NOT_TRUE			2

/* iSCSI Drafts */
#define DRAFT8				800
#define DRAFT9				900
#define DRAFT11				1100
#define DRAFT12				1200
#define DRAFT13				1300
#define DRAFT18				1800
#define DRAFT_MULTIPLIER		100

#define DEFAULT_DRAFT			DRAFT18


/* iSCSI version */
#define ISCSI_MAX_VERSION		3	/* Draft 9 */
#define ISCSI_MIN_VERSION		0	/* Draft 11,12,13,18 */

/********************************************************************/
/* Bits used in the "neg_info" variable and key field during Login Phase */

/* these 2 bits:
	are passed by manage in the neg_info parameter to proc_info()
	are used by proc_info()
	are never stored in param->neg_info
	are never used during initiator or target operation
		(i.e., they are only used between manage and proc_info())
*/
#define RESTORE 		0x00000001
#define FORCE			0x00000002

/* these 2 bits:
	are passed by manage in the neg_info parameter to proc_info()
	are used by proc_info()
	are stored by proc_info() directly in param->neg_info
	are tested in param->neg_info during login negotiations
	are never used in FFP
*/
#define KEY_TO_BE_NEGOTIATED  	0x00000004 
#define KEY_BREAK_CONN_OR_RJT	0x00000008

/* these 6 bits:
	are not used by manage
	are not used by proc_info()
	are set and tested in param->neg_info during login negotiations
	are never used in FFP
*/
#define KEY_SENT_TO_OTHER_SIDE	0x00000010
#define KEY_GOT_FROM_OTHER_SIDE	0x00000020
#define KEY_REPLY_OPTIONAL	0x00000040
#define KEY_IRRELEVANT		0x00000080
#define KEY_BAD			0x00000100
#define KEY_REJECT		0x00000200
#define KEY_WRONG		0x00000400
/********************************************************************/


/* iSCSI opcodes */

#define ISCSI_INIT_NOP_OUT			0x00
#define ISCSI_INIT_SCSI_CMND			0x01
#define ISCSI_INIT_TASK_MGMT_CMND		0x02
#define ISCSI_INIT_LOGIN_CMND			0x03
#define ISCSI_INIT_TEXT_CMND			0x04
#define ISCSI_INIT_SCSI_DATA_OUT		0x05
#define ISCSI_INIT_LOGOUT_CMND			0x06
#define ISCSI_INIT_SNACK			0x10

#define ISCSI_TARG_NOP_IN			0x20
#define ISCSI_TARG_SCSI_RSP			0x21
#define ISCSI_TARG_TASK_MGMT_RSP		0x22
#define ISCSI_TARG_LOGIN_RSP			0x23
#define ISCSI_TARG_TEXT_RSP			0x24
#define ISCSI_TARG_SCSI_DATA_IN			0x25
#define ISCSI_TARG_LOGOUT_RSP			0x26
#define ISCSI_TARG_R2T				0x31
#define ISCSI_TARG_ASYNC_MSG			0x32
#define ISCSI_TARG_RJT				0x3f

/* bits in connection_flags to control login and FFP */
#define USE_FLAT_SPACE_LUN			0x0001
#define USE_FULL_REPLIES			0x0002
#define USE_SECURITY_PHASE			0x0004
#define USE_OPERATIONAL_PHASE			0x0008
#define USE_HEADERDIGEST			0x0010
#define USE_DATADIGEST				0x0020
#define GOT_ACTIVITY				0x0040
#define SEND_NO_REPLY_TO_NOP_PING		0x0080	/* test 6 */
#define SEND_NO_REPLY_TO_ASYNC_LOGOUT		0x0100	/* test 7 */

/* chap support - CHONG */
#define USE_TARGET_CONFIRMATION                 0x01
#define USE_BASE64                              0x02

/* bits and fields in the opcode byte (always byte 0 in a pdu) */
#define ISCSI_OPCODE				0x3f
#define REPLY_BIT				0x20
#define I_BIT                           	0x40
#define OLD_X_BIT				0x80	/* draft 8 and before */

/* bits and fields in the flags byte (always byte 1 in a pdu) */
#define F_BIT                  			0x80
#define R_BIT					0x40
#define A_BIT					0x40
#define C_BIT					0x40
#define READ_TYPE_CMND         			R_BIT
#define W_BIT					0x20
#define WRITE_TYPE_CMND        			W_BIT
#define BRO_BIT					0x10
#define BRU_BIT					0x08
#define O_BIT					0x04
#define U_BIT					0x02
#define S_BIT                           	0x01

#define T_BIT                                   0x80
#define CSG                                     0x0C
#define CSG1					0x04
#define CSG2					0x08
#define CSG3                                    0x0C
#define CSG_SHIFT				2
#define NSG                                     0x03
#define NSG1					0x01
#define NSG2					0x02
#define NSG3                                    0x03



/* SCSI Command ATTR value */
#define ISCSI_UNTAGGED         			0x00
#define ISCSI_SIMPLE   		       		0x01
#define ISCSI_ORDERED          			0x02
#define ISCSI_HEAD_OF_QUEUE    			0x03
#define ISCSI_ACA              			0x04
#define ISCSI_STATUS           			0x04


#ifndef __SCSI_TARGET__

/* task management function values */
#define ABORT_TASK            		1
#define ABORT_TASK_SET        		2
#define CLEAR_ACA             		3
#define CLEAR_TASK_SET        		4
#define LUN_RESET             		5
#define TARGET_WARM_RESET     		6
#define TARGET_COLD_RESET     		7
#define TARGET_REASSIGN				8
#define TASK_REASSIGN				8 /* Changed to Task - SAI */

#endif

#define ISCSI_SHUTDOWN_SIGS    (sigmask(SIGKILL)|sigmask(SIGINT)|sigmask(SIGTERM))

/* for status class field in login response */
#define STAT_CLASS_SUCCESS			0x00
#define STAT_CLASS_REDIRECTION			0x01
#define STAT_CLASS_INITIATOR			0x02
#define STAT_CLASS_TARGET			0x03

#define STAT_DETAIL_ERR 			0x00
#define STAT_DETAIL_TARG_MOVED_TEMP 		0x01
#define STAT_DETAIL_TARG_MOVED_PERM 		0x02
#define STAT_DETAIL_NOT_AUTH 			0x01
#define STAT_DETAIL_NOT_ALLOWED 		0x02
#define STAT_DETAIL_NOT_FOUND 			0x03
#define STAT_DETAIL_TARG_REMOVED 		0x04
#define STAT_DETAIL_VERSION_NOT_SUPPORTED 	0x05
#define STAT_DETAIL_TOO_MANY_CONNECTIONS 	0x06
#define STAT_DETAIL_MISSING_PARAMETER 		0x07
#define STAT_DETAIL_NOT_INCLUDED 		0x08
#define STAT_DETAIL_SESSION_TYPE 		0x09
#define STAT_DETAIL_SESSION_NOT_EXIST 		0x0a
#define STAT_DETAIL_INVALID_DURING_LOGIN	0x0b
#define STAT_DETAIL_SERVICE_UNAVAILABLE 	0x01
#define STAT_DETAIL_OUT_OF_RESOURCE 		0x02


/* task managament response values */
#define FUNCTION_COMPLETE           	0
#define TASK_WAS_NOT_IN_TASK_SET    	1
#define LUN_DOES_NOT_EXIST          	2
#define TASK_STILL_ALLEGIANT		3
#define TASK_FAILOVER_NOT_SUPPORTED	4
#define FUNCTION_REJECTED           	255

/* values for "state" in struct iscsi_cmnd */
#define ISCSI_CMND_RECEIVED            	1
#define ISCSI_NEW_CMND                  2
#define ISCSI_BUFFER_RDY                3
#define ISCSI_DONE                      4
#define ISCSI_SENT                      5
#define ISCSI_DEQUEUE                   6
#define ISCSI_R2T_SENT                  7
#define ISCSI_IMMEDIATE_DATA_IN         8
#define ISCSI_UNSOLICITED_DATA_IN       9
#define ISCSI_DATA_IN                   10
#define ISCSI_MGT_FN_DONE               11
#define ISCSI_DISCOVERY			12
#define ISCSI_LOGOUT			13
#define ISCSI_PING			14

/* reason code in Reject */
#define REASON_FULL_BEFORE_LOGIN		0x01
#define REASON_DATA_DIGEST_ERR			0x02
#define REASON_DATA_SNACK			0x03
#define REASON_PROTOCOL_ERR			0x04
#define REASON_COMMAND_NOT_SUPPORTED		0x05
#define	REASON_TOO_MANY_IMMEDIATE_COMMANDS	0x06
#define REASON_TASK_IN_PROGRESS			0x07
#define REASON_INVALID_SNACK			0x08
#define REASON_NO_BOOKMARK			0x09
#define REASON_OUT_OF_RESOURCES			0x0a
#define REASON_NEGOTIATION_RESET		0x0b

/* reason code in logout - SAI */
#define LOGOUT_REASON				0x7f

/* For Text Parameter Negotiation */
/* Draft 18, Section 4.3 Login Phase
   "The default MaxRecvDataSegmentLength is used during Login."
   Draft 18, Section 11.12 MaxRecvDataSegmentLength
   "Default is 8192 bytes."
*/
#define MAX_TEXT_LEN 			8192

/* SNACK for Error Recovery - SAI */
#define SNACK                           0x00000030

/* iSCSI message formats */

struct iscsi_init_scsi_cmnd		/* Draft 18, Section 9.3 SCSI Command */
{
	__u8	opcode;
	__u8	flags;
	__u8	rsvd1;
	__u8	crn;			/* only used in drafts < 12 */
	__u32	length;
	__u64	lun;
	__u32	init_task_tag;
	__u32 	xfer_len;
	__u32	cmd_sn;
	__u32	exp_stat_sn;
	__u8	cdb[16];
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_targ_scsi_rsp		/* Draft 18, Section 9.4 SCSI Response */
{
	__u8	opcode;
	__u8	flags;
	__u8	response;
	__u8	status;
	__u32	length;
	__u64	lun;			/* reserved */
	__u32	init_task_tag;
	__u32	target_xfer_tag;	/* ResidualCount in drafts <= 9 */
	__u32	stat_sn;
	__u32	exp_cmd_sn;
	__u32	max_cmd_sn;
	__u32	exp_data_sn;
	__u32	bidi_resid;		/* reserved in drafts <= 9 */
	__u32	resid;			/* bidi_resid in drafts <= 9 */
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_init_task_mgt_command	/* Draft 18, Section 9.5 Task Management
					   Function Request */
{
    	__u8    opcode;
    	__u8    function;
    	__u16   rsvd1;
    	__u32   length;			/* reserved */
    	__u64   lun;
    	__u32   init_task_tag;
    	__u32   ref_task_tag;
    	__u32   cmd_sn;
    	__u32   exp_stat_sn;
    	__u32	ref_cmd_sn;
    	__u32   exp_data_sn;		/* reserved in drafts < 12 */
    	__u64   rsvd4;
    	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_targ_task_mgt_response	/* Draft 18, Section 9.6
					   Task Management Function Response */
{
    	__u8    opcode;
    	__u8    flags;
    	__u8	response;
    	__u8	qualifier;		/* reserved in drafts >= 11 */
    	__u32	length;			/* reserved */
    	__u64	lun;			/* reserved */
    	__u32   init_task_tag;
    	__u32   ref_task_tag;		/* reserved in drafts >= 12 */
    	__u32   stat_sn;
    	__u32   exp_cmd_sn;
    	__u32   max_cmd_sn;
    	__u32   rsvd4;
    	__u64   rsvd5;
    	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_init_scsi_data_out		/* Draft 18, Section 9.7 SCSI Data-out*/
{
	__u8	opcode;
	__u8	flags;
	__u16	rsvd2;
	__u32	length;
	__u64	lun;
	__u32	init_task_tag; 
	__u32	target_xfer_tag;
	__u32	rsvd3;
	__u32	exp_stat_sn;
	__u32	rsvd4;
	__u32	data_sn;
	__u32	offset;
	__u32	rsvd5;
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_targ_scsi_data_in		/* Draft 18, Section 9.7 SCSI Data-in */
{
	__u8	opcode;
	__u8	flags;
	__u8	rsvd1;
	__u8	status;
	__u32	length;
	__u64	lun;			/* reserved in drafts <= 9 */
	__u32	init_task_tag; 
	__u32	target_xfer_tag;	/* ResidualCount in drafts <= 9 */
	__u32	stat_sn;
	__u32	exp_cmd_sn;
	__u32	max_cmd_sn;
	__u32	data_sn;
	__u32	offset;
	__u32	resid;			/* reserved in drafts <= 9 */
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_targ_r2t			/* Draft 18, Section 9.8
					   Ready To Transfer (R2T)*/
{
	__u8	opcode;
	__u8	flags;
	__u16	rsvd2;
	__u32	length;			/* reserved */
	__u64	lun;			/* reserved in drafts <= 9 */
	__u32	init_task_tag; 
	__u32	target_xfer_tag;
	__u32	stat_sn;
	__u32	exp_cmd_sn;
	__u32	max_cmd_sn;
	__u32	r2t_sn;
	__u32	offset;
	__u32	xfer_len;
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_targ_async_msg		/* Draft 18, Section 9.9
					   Asynchronous Message */
{
	__u8	opcode;
	__u8	flags;
	__u16	rsvd2;
	__u32	length;
	__u64	lun;
	__u32	init_task_tag; 		/* reserved 0xffffffff */
	__u32	rsvd3;
	__u32	stat_sn;
	__u32	exp_cmd_sn;
	__u32	max_cmd_sn;
	__u8	async_event;
	__u8	async_vcode;
	__u16	parameter1;
	__u16	parameter2;
	__u16	parameter3;
	__u32	rsvd5;
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_init_text_cmnd		/* Draft 18, Section 9.10 Text Request*/
{
	__u8	opcode;
	__u8	flags;
	__u16	rsvd2; 
	__u32	length;
	__u64	lun;			/* reserved in drafts <= 9 */
	__u32	init_task_tag;
	__u32	target_xfer_tag;
	__u32	cmd_sn;
	__u32	exp_stat_sn;
	__u64	rsvd4;
	__u64	rsvd5;
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_targ_text_rsp		/*Draft 18, Section 9.11 Text Response*/
{
	__u8	opcode;
	__u8	flags;
	__u16	rsvd2; 
	__u32	length;
	__u64	lun;			/* reserved in drafts <= 9 */
	__u32	init_task_tag;
	__u32	target_xfer_tag;
	__u32	stat_sn;
	__u32	exp_cmd_sn;
	__u32	max_cmd_sn;
	__u32	rsvd4;
	__u64	rsvd5;
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_init_login_cmnd 		/*Draft 18, Section 9.12 Login Request*/
{
	__u8 	opcode;
	__u8	flags;
	__u8 	version_max;
	__u8	version_min;
	__u32	length;
	__u8	isid[6];
	__u16	tsih;
	__u32 	init_task_tag;
	__u16	cid;
	__u16	rsvd1;
	__u32	cmd_sn;
	__u32	exp_stat_sn;
	__u64	rsvd2;
	__u64	rsvd3;
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_targ_login_rsp		/* Draft 18, Section 9.13
					   Login Response */
{
	__u8 	opcode;
	__u8	flags;
	__u8	version_max;
	__u8	version_active;
	__u32	length;
	__u8	isid[6];
	__u16	tsih;
	__u32	init_task_tag;
	__u32	rsvd1;
	__u32	stat_sn;
	__u32	exp_cmd_sn;
	__u32	max_cmd_sn;
	__u8	status_class;
	__u8	status_detail;
	__u16	rsvd2;
	__u64	rsvd3;
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_init_logout_cmnd		/* Draft 18, Section 9.14
					   Logout Request */
{
	__u8	opcode;
	__u8	flags;			/* reasoncode in Draft 13 and after */
	__u16	rsvd1; 
	__u32	length;
	__u64	lun;			/* reserved */
	__u32	init_task_tag;
	__u16	cid;
	__u8	rsvd2;
	__u8	rsvd3;			/* reasoncode before Draft 13 */
	__u32	cmd_sn;
	__u32	exp_stat_sn;
	__u64	rsvd4;
	__u64	rsvd5;
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_targ_logout_rsp		/* Draft 18, Section 9.15
					   Logout Response */
{
	__u8	opcode;
	__u8	flags;
	__u8	response;
	__u8	rsvd1; 
	__u32	length;			/* reserved */
	__u64	lun;			/* reserved */
	__u32	init_task_tag;
	__u32	rsvd3;
	__u32	stat_sn;
	__u32	exp_cmd_sn;
	__u32	max_cmd_sn;
	__u32	rsvd4;
	__u16	time2wait;
	__u16	time2retain;
	__u32	rsvd5;
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_targ_rjt			/* Draft 18, Section 9.17 Reject */
{
	__u8 	opcode;
	__u8	flags;
	__u8	reason;
	__u8	rsvd2;
	__u32	length;
	__u64	lun;			/* reserved */
	__u32	init_task_tag;		/* reserved 0xffffffff */
	__u32	rsvd4;
	__u32	stat_sn;
	__u32	exp_cmd_sn;
	__u32	max_cmd_sn;
	__u32	data_sn;
	__u64	rsvd5;
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_init_nopout		/* Draft 18, Section 9.18 NOP-Out */
{
	__u8 	opcode;
	__u8	flags;
	__u16	rsvd1;
	__u32	length;
	__u64	lun;
	__u32	init_task_tag;
	__u32	target_xfer_tag;
	__u32	cmd_sn;
	__u32	exp_stat_sn;
	__u64	rsvd2;
	__u64	rsvd3;
	__u32	header_digest;		/* holds header CRC if in use */
};

struct iscsi_targ_nopin			/* Draft 18, Section 9.19 NOP-In */
{
	__u8 	opcode;
	__u8	flags;
	__u16	rsvd1;
	__u32	length;
	__u64	lun;
	__u32	init_task_tag;
	__u32	target_xfer_tag;
	__u32	stat_sn;
	__u32	exp_cmd_sn;
	__u32	max_cmd_sn;
	__u32	rsvd2;
	__u64	rsvd3;
	__u32	header_digest;		/* holds header CRC if in use */
};

/* Used in Login negotiations */
struct generic_pdu
{
	__u8	opcode;
	__u8	flags;
	__u8	version_max;
	__u8	version_active;		/* for login response PDU */
	__u32	length;
	__u8	isid[6];
	__u16	tsih;
	__u32 	init_task_tag;
	__u16	CID;
	__u16	rsvd1;
	__u32	cmd_sn;
	__u32 	exp_stat_sn;
	__u32   max_cmd_sn;
	__u8	status_class;
	__u8	status_detail;
	__u16	rsvd2;
	__u64	rsvd3;
	__u32	header_digest;		/* holds header CRC if in use */
	__u8	*text;
	__u32	text_length;
};

/* Used in generic processing of responses */
struct response_pdu
{
	__u8	opcode;
	__u8	flags;
	__u8	reason;			/* for reject PDU */
	__u8	version_active;		/* for login response PDU */
	__u32	length;
	__u64	lun;
	__u32 	init_task_tag;
	__u32	target_xfer_tag;
	__u32	stat_sn;
	__u32	exp_cmd_sn;
	__u32	max_cmd_sn;
	__u8	status_class;
	__u8	status_detail;
	__u16	rsvd2;
	__u32	offset;
	__u32	xfer_len;
	__u32	header_digest;		/* holds header CRC if in use */
};

/* SNACK pdu structure definition - SAI */

struct iscsi_init_snack
        {
        __u8    opcode;
        __u8    flags;
        __u16   rsvd1;
        __u32   length;
        __u64   lun;
        __u32   init_task_tag;
        __u32   target_xfer_tag;
        __u32   rsvd2;
        __u32   exp_stat_sn;
        __u64   rsvd3;
        __u32   begrun;
        __u32   runlen;
        __u32   header_digest;
        };

/* SENSE Data structure definition - SAI */

struct scsi_sense_data
        {
        __u8    response;
        __u8    obs;
        __u8    sense_key;
        __u16   info;
        __u8    asl;
        __u16   cmnd_spec_info;
        __u8    asc;
        __u8    ascq;
        __u8    field;
        __u16   sense_specific;
        };

/* values of connection-specific parameters used during FFP */
struct connection_operational_parameters
{
	void	*dummy;			/* at present, there are none we use */
};

/* values of session-wide parameters used during FFP */
struct session_operational_parameters
{
	__u16	MaxConnections;			/* [1..65535] */
	__u8	InitialR2T;			/* [0,1] == [No,Yes] */
	__u8	BidiInitialR2T;			/* [0,1] == [No,Yes] */
	__u8	ImmediateData;			/* [0,1] == [No,Yes] */
	__u32	DataPDULength;			/* [512..2**24-1] */
	//__u32	MaxRecvPDULength;		/* [512..2**24-1] */
	//__u32	MaxRecvDataSegmentLength;	/* [512..2**24-1] */
	__u32	MaxBurstSize;			/* [512..2**24-1] */
	__u32 	FirstBurstSize;			/* [512..2**24-1] */
	__u32	MaxBurstLength;			/* [512..2**24-1] */
	__u32 	FirstBurstLength;		/* [512..2**24-1] */
	__u16	DefaultTime2Wait;		/* [0..3600] */
	__u16	LogoutLoginMinTime;		/* [2..3600] */
	__u16	DefaultTime2Retain;		/* [0..3600] */
	__u16	LogoutLoginMaxTime;		/* [2..3600] */
	__u16	MaxOutstandingR2T;		/* [1..65535] */
	__u8	DataPDUInOrder;			/* [0,1] == [No,Yes] */
	__u8	DataSequenceInOrder;		/* [0,1] == [No,Yes] */
	__u8	ErrorRecoveryLevel;		/* [0..2] */
	__u8	SessionType;			/* [0,1] == [Normal,Discovery]*/
	__u8    TargetName;                     /* The number at the end of the target name */
};

#ifdef __KERNEL__

char *string_llx( unsigned long long x )
{
    int			i;
    static char		str[20];
    static char		convert[] = "0123456789abcdef";
    char		*ptr;
    unsigned int	c;
    union {
    	  unsigned long long	x;
	  unsigned char		c[8];
	  }	tmp;

    tmp.x = x;
    ptr = str;
    *ptr++ = '0';
    *ptr++ = 'x';
    for( i = 7;  i >= 0;  i-- )
    	{
	c = tmp.c[i];
	*ptr++ = convert[c>>4];
	*ptr++ = convert[c&0xf];
	}
    *ptr = '\0';
    return str;
}

void print_rsvd_u8( int n, unsigned char rsvd )
{
  if( rsvd != 0 )
    printk("    rsvd%d: 0x%.2x\n", n, rsvd);
}
void print_rsvd_u16( int n, unsigned short rsvd )
{
  if( rsvd != 0 )
    printk("    rsvd%d: 0x%.4x\n", n, rsvd);
}
void print_rsvd_u32( int n, unsigned int rsvd )
{
  if( rsvd != 0 )
    printk("    rsvd%d: 0x%.8x\n", n, rsvd);
}
void print_rsvd_u64( int n, unsigned long long rsvd )
{
    if( rsvd != 0ll )
      {
      printk("    rsvd%d: %s\n", n, string_llx(rsvd));
      }
}
void print_opcode( unsigned char opcode )
{
    printk("    Opcode: 0x%.2x,  I: %d\n", opcode&0x3f, (opcode&0x40)!=0 );
}
void print_flags( unsigned char flags )
{
    printk("    flags: 0x%.2x\n", flags);
}
void print_lun( unsigned long long lun )
{
    if( lun != 0ll )
      {
      printk("    LUN: %s\n", string_llx(lun));
      }
}
void print_isid_tsih( __u8 isid[6], __u16 tsih )
{
    printk("    ISID: 0x%.2x %.2x %.2x %.2x %.2x %.2x\n",
			isid[0], isid[1], isid[2], isid[3], isid[4], isid[5]);
	/*	need to go back and fix this to print TSID for drafts < 12 */
    printk("    TSIH: %u\n", be16_to_cpu(tsih));
}
void print_dsl( unsigned int length )
{
    if( length != 0 )
      {
      printk("    DSL: %u\n", be32_to_cpu(length));
      }
}
void print_itt( unsigned int init_task_tag )
{
	if( init_task_tag == 0xffffffff )
      printk("    ITT: 0x%08x\n", init_task_tag);
	else
      printk("    ITT: %u\n", be32_to_cpu(init_task_tag));
}
void print_ttt( unsigned int target_xfer_tag )
{
	if( target_xfer_tag == 0xffffffff )
      printk("    TTT: 0x%08x\n", target_xfer_tag);
	else
      printk("    TTT: %u\n", be32_to_cpu(target_xfer_tag));
}
void print_expstatsn( unsigned int exp_stat_sn )
{
    if( exp_stat_sn != 0 )
      {
      printk("    ExpStatSN: %u\n", be32_to_cpu(exp_stat_sn));
      }
}
void print_cmdsn_expstatsn( unsigned int cmd_sn, unsigned int exp_stat_sn )
{
    printk("    CmdSN: %u\n", be32_to_cpu(cmd_sn));
    print_expstatsn(exp_stat_sn);
}
void print_statsn_exp_max( unsigned int stat_sn, unsigned int exp_cmd_sn,
						unsigned int max_cmd_sn )
{
    if( stat_sn != 0 )
      printk("    StatSN: %u\n", be32_to_cpu(stat_sn));
    printk("    ExpCmdSN: %u\n", be32_to_cpu(exp_cmd_sn));
    printk("    MaxCmdSN: %u\n", be32_to_cpu(max_cmd_sn));
}
void print_residual( unsigned int resid )
{
    if( resid != 0 )
      {
      printk("    ResidualCount: %u\n", be32_to_cpu(resid));
      }
}
void print_datasn( unsigned int data_sn )
{
    if( data_sn != 0 )
      {
      printk("    DataSN: %u\n", be32_to_cpu(data_sn));
      }
}
void print_offset( unsigned int offset )
{
    if( offset != 0 )
      {
      printk("    BufferOffset: %u\n", be32_to_cpu(offset));
      }
}
void print_rtt( unsigned int ref_task_tag )
{
	if( ref_task_tag != 0 )
      printk("    RTT: 0x%.8x\n", be32_to_cpu(ref_task_tag));
}
void print_exp_data_sn( unsigned int exp_data_sn )
{
    if( exp_data_sn != 0 )
      printk("    ExpDataSN: %u\n", be32_to_cpu(exp_data_sn));
}

void print_init_scsi_cmnd(struct iscsi_init_scsi_cmnd *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u8(1, cmd->rsvd1);
    /*	need to go back and fix this to print CRN only for drafts < 12 */
    if( cmd->crn != 0 )
      printk("    CRN: 0x%.2x\n", cmd->crn);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    printk("    EDTL: %u\n", be32_to_cpu(cmd->xfer_len));
    print_cmdsn_expstatsn(cmd->cmd_sn ,cmd->exp_stat_sn);
    printk("    CDB: 0x%.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x\n",
		cmd->cdb[0], cmd->cdb[1], cmd->cdb[2], cmd->cdb[3], 
		cmd->cdb[4], cmd->cdb[5], cmd->cdb[6], cmd->cdb[7], 
		cmd->cdb[8], cmd->cdb[9], cmd->cdb[10], cmd->cdb[11],
		cmd->cdb[12], cmd->cdb[13], cmd->cdb[14], cmd->cdb[15]);
}

void print_targ_scsi_rsp(struct iscsi_targ_scsi_rsp *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    if( cmd->response != 0 )
      printk("    Response: 0x%.2x\n", cmd->response);
    if( cmd->status != 0 )
      printk("    Status: 0x%.2x\n", cmd->status);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_exp_data_sn(cmd->exp_data_sn);
    if( cmd->bidi_resid != 0 )
      printk("    BidiResidualCount: %u\n", be32_to_cpu(cmd->bidi_resid));
    print_residual(cmd->resid);
}

void print_init_text_cmnd(struct iscsi_init_text_cmnd *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_cmdsn_expstatsn(cmd->cmd_sn ,cmd->exp_stat_sn);
    print_rsvd_u64(4, cmd->rsvd4);
    print_rsvd_u64(5, cmd->rsvd5);
}

void print_targ_text_rsp(struct iscsi_targ_text_rsp *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_rsvd_u32(4, cmd->rsvd4);
    print_rsvd_u64(5, cmd->rsvd5);
}

void print_init_login_cmnd(struct iscsi_init_login_cmnd *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    printk("    VersionMax: 0x%.2x\n", cmd->version_max);
    printk("    VersionMin: 0x%.2x\n", cmd->version_min);
    print_dsl(cmd->length);
    print_isid_tsih(cmd->isid, cmd->tsih);
    print_itt(cmd->init_task_tag);
    printk("    CID: %u\n", be16_to_cpu(cmd->cid));
    print_rsvd_u16(1, cmd->rsvd1);
    print_cmdsn_expstatsn(cmd->cmd_sn ,cmd->exp_stat_sn);
    print_rsvd_u64(2, cmd->rsvd2);
    print_rsvd_u64(3, cmd->rsvd3);
}

void print_targ_login_rsp(struct iscsi_targ_login_rsp *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    printk("    VersionMax: 0x%.2x\n", cmd->version_max);
    printk("    VersionActive: 0x%.2x\n", cmd->version_active);
    print_dsl(cmd->length);
    print_isid_tsih(cmd->isid, cmd->tsih);
    print_itt(cmd->init_task_tag);
    print_rsvd_u32(1, cmd->rsvd1);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    if( cmd->status_class != 0 )
      printk("    StatusClass: 0x%.2x\n", cmd->status_class);
    if( cmd->status_detail != 0 )
      printk("    StatusDetail: 0x%.2x\n", cmd->status_detail);
    print_rsvd_u16(2, cmd->rsvd2);
    print_rsvd_u64(3, cmd->rsvd3);
}

void print_init_logout_cmnd(struct iscsi_init_logout_cmnd *cmd)
{
    print_opcode(cmd->opcode);
    printk("reasoncod: 0x%.2x\n", cmd->flags);
    print_rsvd_u16(1, cmd->rsvd1);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    if( cmd->cid != 0 )
      printk("    CID: %u\n", be16_to_cpu(cmd->cid));
    print_rsvd_u8(2, cmd->rsvd2);
    print_rsvd_u8(3, cmd->rsvd3);
    print_cmdsn_expstatsn(cmd->cmd_sn ,cmd->exp_stat_sn);
    print_rsvd_u64(4, cmd->rsvd4);
    print_rsvd_u64(5, cmd->rsvd5);
}

void print_targ_logout_rsp(struct iscsi_targ_logout_rsp *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    printk(" response: 0x%.2x\n", cmd->response);
    print_rsvd_u8(1, cmd->rsvd1);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_rsvd_u32(3, cmd->rsvd3);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_rsvd_u32(4, cmd->rsvd4);
    printk("Time2Wait: 0x%.8x\n", be16_to_cpu(cmd->time2wait));
    printk("Tm2Retain: 0x%.8x\n", be16_to_cpu(cmd->time2retain));
    print_rsvd_u32(5, cmd->rsvd5);
}

void print_init_scsi_data_out(struct iscsi_init_scsi_data_out *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_rsvd_u32(3, cmd->rsvd3);
    print_expstatsn(cmd->exp_stat_sn);
    print_rsvd_u32(4, cmd->rsvd4);
    print_datasn(cmd->data_sn);
    print_offset(cmd->offset);
    print_rsvd_u32(5, cmd->rsvd5);
}

void print_targ_scsi_data_in(struct iscsi_targ_scsi_data_in *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u8(1, cmd->rsvd1);
    if( cmd->status != 0 )
      printk("    Status: 0x%.2x\n", cmd->status);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_datasn(cmd->data_sn);
    print_offset(cmd->offset);
    print_residual(cmd->resid);
}

void print_targ_rjt(struct iscsi_targ_rjt *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    if( cmd->reason != 0 )
      printk("    Reason: 0x%.2x\n", cmd->reason);
    print_rsvd_u8(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_rsvd_u32(4, cmd->rsvd4);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_datasn(cmd->data_sn);
    print_rsvd_u64(4, cmd->rsvd4);
    print_rsvd_u64(5, cmd->rsvd5);
}

void print_init_nopout(struct iscsi_init_nopout *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(1, cmd->rsvd1);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_cmdsn_expstatsn(cmd->cmd_sn ,cmd->exp_stat_sn);
    print_rsvd_u64(2, cmd->rsvd2);
    print_rsvd_u64(3, cmd->rsvd3);
}

void print_targ_nopin(struct iscsi_targ_nopin *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(1, cmd->rsvd1);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_statsn_exp_max(cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_rsvd_u32(2, cmd->rsvd2);
    print_rsvd_u64(3, cmd->rsvd3);
}

void print_targ_r2t(struct iscsi_targ_r2t *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_ttt(cmd->target_xfer_tag);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    printk("    R2TSN: %u\n", be32_to_cpu(cmd->r2t_sn));
    print_offset(cmd->offset);
    printk("    DDTL: %u\n", be32_to_cpu(cmd->xfer_len));
}

void print_targ_async_msg(struct iscsi_targ_async_msg *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    print_rsvd_u16(2, cmd->rsvd2);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_rsvd_u32(3, cmd->rsvd3);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    printk("AsyncEvnt: %u\n", cmd->async_event);
    printk("AsyncVCod: %u\n", cmd->async_vcode);
    if( cmd->parameter1 != 0 )
	printk("   Param1: %u\n", cmd->parameter1);
    if( cmd->parameter2 != 0 )
	printk("   Param2: %u\n", cmd->parameter2);
    if( cmd->parameter3 != 0 )
	printk("   Param3: %u\n", cmd->parameter3);
    print_rsvd_u32(5, cmd->rsvd5);
}

void print_init_task_mgt_command(struct iscsi_init_task_mgt_command *cmd)
{
    print_opcode(cmd->opcode);
    printk("    Function: 0x%.2x\n", cmd->function);
    print_rsvd_u16(1, cmd->rsvd1);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_rtt(cmd->ref_task_tag);
    print_cmdsn_expstatsn(cmd->cmd_sn, cmd->exp_stat_sn);
    if( cmd->ref_cmd_sn != 0 )
      printk("    RefCmdSN: %u\n", be32_to_cpu(cmd->ref_cmd_sn));
    print_exp_data_sn(cmd->exp_data_sn);
    print_rsvd_u64(4, cmd->rsvd4);
}

void print_targ_task_mgt_response(struct iscsi_targ_task_mgt_response *cmd)
{
    print_opcode(cmd->opcode);
    print_flags(cmd->flags);
    if( cmd->response != 0 )
      printk("    Response: 0x%.2x\n", cmd->response);
    if( cmd->qualifier != 0 )
      printk("    Qualifier: 0x%.2x\n", cmd->qualifier);
    print_dsl(cmd->length);
    print_lun(cmd->lun);
    print_itt(cmd->init_task_tag);
    print_rtt(cmd->ref_task_tag);
    print_statsn_exp_max( cmd->stat_sn , cmd->exp_cmd_sn, cmd->max_cmd_sn);
    print_rsvd_u32(4, cmd->rsvd4);
    print_rsvd_u64(5, cmd->rsvd5);
}

/* Print SNACK Request - SAI */

/********************************************************************************
*       Called to print the SNACK Request being sent
********************************************************************************/

void print_begrun(unsigned int begrun)
        {
        printk("    BegRun: %u\n", be32_to_cpu(begrun));
        }

void print_runlen(unsigned int runlen)
        {
        printk("    RunLength: %u\n", be32_to_cpu(runlen));
        }

void print_init_snack(struct iscsi_init_snack *cmd)
        {
        print_opcode(cmd->opcode);
        print_flags(cmd->flags);
        print_rsvd_u16(1, cmd->rsvd1);
        print_dsl(cmd->length);
        print_lun(cmd->lun);
        print_itt(cmd->init_task_tag);
        print_ttt(cmd->target_xfer_tag);
        print_rsvd_u32(2, cmd->rsvd2);
        print_expstatsn(cmd->exp_stat_sn);
        print_rsvd_u64(3, cmd->rsvd3);
        print_begrun(cmd->begrun);
        print_runlen(cmd->runlen);
        }

/*	returns pointer to printable name of SCSI op code */
/*	(Normally this is byte 1 of a SCSI CDB */
char *printable_scsi_op( unsigned int opcode )
	{
	static char		buffer[16];
	char			*result = "";


	/*	opcode symbols defined in <scsi/scsi.h> */
	/*	this table includes only those symbols we have actually seen
		being sent to/from a vendor */
	switch( opcode )
		{
	case TEST_UNIT_READY:		/* 0x00 */
		result = "TEST_UNIT_READY";
		break;
	case REZERO_UNIT:			/* 0x01 */
		result = "REWIND";
		break;
	case REQUEST_SENSE:			/* 0x03 */
		result = "REQUEST_SENSE";
		break;
	case READ_BLOCK_LIMITS:		/* 0x05 */
		result = "READ_BLOCK_LIMITS";
		break;
	case READ_6:				/* 0x08 */
		result = "READ_6";
		break;
	case WRITE_6:				/* 0x0a */
		result = "WRITE_6";
		break;
	case WRITE_FILEMARKS:		/* 0x10 */
		result = "WRITE_FILEMARKS";
		break;
	case INQUIRY:				/* 0x12 */
		result = "INQUIRY";
		break;
	case MODE_SENSE:			/* 0x1a */
		result = "MODE_SENSE";
		break;
	case READ_CAPACITY:			/* 0x25 */
		result = "READ_CAPACITY";
		break;
	case READ_10:				/* 0x28 */
		result = "READ_10";
		break;
	case WRITE_10:				/* 0x2a */
		result = "WRITE_10";
		break;
	case READ_12:				/* 0xa8 */
		result = "READ_12";
		break;
	case WRITE_12:				/* 0xaa */
		result = "WRITE_12";
		break;
	default:
		sprintf(buffer, "CDBop 0x%02x", opcode);
		result = buffer;
		break;
		}	/* switch */
	return result;
	}

#include "my_memory.h"

#endif

#endif
