
#define MEMORY_TABLE_SIZE	512

void	*my_memory_address_table[MEMORY_TABLE_SIZE];
char	*my_memory_id_table[MEMORY_TABLE_SIZE];
int		my_memory_last = -1;		/* currently the last used slot */
int		my_memory_first = 0;		/* currently first possible unused slot */
int		my_memory_max = -1;			/* maximum last used slot so far */

/*	called to allocate memory and keep a record of it */
void *my_kmalloc( unsigned int size, char *id )
	{
	void	*ptr;
	int		i;


	TRACE((TRACE_ENTER_LEAVE|TRACE_MY_MEMORY),
	"Enter my_kmalloc %u bytes, last %d, for %s\n", size, my_memory_last, id);

	if( (ptr = kmalloc(size, GFP_KERNEL)) == NULL )
		{
		TRACE_ERROR("Cannot allocate %u bytes for %s\n", size, id);
		goto out;
		}

	/*	search for an unused slot, starting with first possible empty one */
	for( i = my_memory_first;  i < MEMORY_TABLE_SIZE;  i++ )
		{
		if( my_memory_address_table[i] == NULL )
			{
			my_memory_address_table[i] = ptr;
			my_memory_id_table[i] = id;
			if( my_memory_last < i )
				{
				my_memory_last = i;
				if( my_memory_max < i )
					my_memory_max = i;
				}
			my_memory_first = i+1;	/* first possible unused slot */
			goto out;
			}
		}

	/* if loop finishes, table is full so we cannot track our allocations */
	TRACE_ERROR("my_memory_table full, %d slots\n", MEMORY_TABLE_SIZE);

out:
	TRACE((TRACE_ENTER_LEAVE|TRACE_MY_MEMORY),
			"Leave my_kmalloc %p, last %d, for %s\n", ptr, my_memory_last, id);

	return ptr;
	}


/*	called to release memory allocated dynamically above */
void my_kfree( void **ptr, char *id )
	{
	int		i;


	TRACE((TRACE_ENTER_LEAVE|TRACE_MY_MEMORY),
			"Enter my_kfree %p, last %d, for %s\n", *ptr, my_memory_last, id);

	if( *ptr != NULL )
		{/* seach table backwards to find slot containing this pointer */
		for( i = my_memory_last;  i >= 0;  i-- )
			{
			if( my_memory_address_table[i] == *ptr )
				{/* found the slot containing this pointer, NULL everything */
				kfree(*ptr);
				*ptr = NULL;
				my_memory_address_table[i] = NULL;
				if( my_memory_first > i )
					my_memory_first = i;
				while( my_memory_last >= 0  &&
							my_memory_address_table[my_memory_last] == NULL )
					{
					if( my_memory_first > my_memory_last )
						my_memory_first = my_memory_last;
					my_memory_last--;
					}
				goto out;
				}
			}

		TRACE_ERROR("trying to free non-malloced memory for %s\n", id);
		}

out:
	TRACE((TRACE_ENTER_LEAVE|TRACE_MY_MEMORY),
					"Leave my_kfree, last %d, for %s\n", my_memory_last, id);
	}


/*	"prints" statistics on my memory usage into buf,
	returns no. of chars printed.
*/
int print_my_kmemory( char *buf )
	{
	if( buf == NULL )
		return printk("iscsi: Maximum of %d memory slots used out of %d\n",
											my_memory_max+1, MEMORY_TABLE_SIZE);
	else
		return sprintf(buf, "Maximum of %d memory slots used out of %d\n",
											my_memory_max+1, MEMORY_TABLE_SIZE);
	}


/*	called only once, at very end when module is unloaded,
	to check for any allocated but not freed memory */
void my_kempty( void )
	{

	TRACE((TRACE_ENTER_LEAVE|TRACE_MY_MEMORY),
								"Enter my_kempty, last %d\n", my_memory_last);

	print_my_kmemory(NULL);

	while( my_memory_last >= 0 )
		{
		if( my_memory_address_table[my_memory_last] != NULL )
			{
			TRACE_ERROR("Unfreed memory[%d] %p for %s\n",
										my_memory_last,
										my_memory_address_table[my_memory_last],
										my_memory_id_table[my_memory_last]);
			/***
			kfree(my_memory_address_table[my_memory_last]);
			***/
			my_memory_address_table[my_memory_last] = NULL;
			}
		my_memory_last--;
		}

	/*	everything should now be reset to initial state */
	my_memory_max = -1;
	my_memory_first = 0;

	TRACE((TRACE_ENTER_LEAVE|TRACE_MY_MEMORY),
								"Leave my_kempty, last %d\n", my_memory_last);
	}
